/**
 * Group2Player2J.java
 * 10/23/2002
 *
 * Improvements to be made:
 *  - take "filled" squares into account when computing 'movesToComplete' and choosing new squares.
 *  - (then make sure to generate a new square instead of simply short-circuiting it)
 *  - exploit existing lines as much as possible when choosing new squares
 *  - switch to a new square when the potential gain of the current one falls below a certain parameter
 *
 * NOTE: Maybe we can assume a robot will continue moving in the direction
 * of its previous move? If this were otherwise a computationally difficult
 * problem, maybe we could do that for simplification.
 */

package Rectangles;

import ui.*;
import java.util.*;
import java.io.*;
import java.awt.Color;

public final class OldGroup2Player5J implements IFCPlayer {

    static final String _CNAME = "G2P5J: Paranoid Player";
    static final Color  _CCOLOR = new Color(0.1f, .7f, 0.1f);
    private static final char[] MOVE = {'E', 'N', 'W', 'S'};
    private static final int X = 0;
    private static final int Y = 1;
    private static final int VERTEX_1 = 0;
    private static final int VERTEX_3 = 1;
    private static final int SHORT_CIRCUIT = 0;
    private static final int GET_NEW = 1;
    private static final int EXTEND = 2;
    private static final int _CFILLED = -1;
    
    private  Rectangles rect;
    private static Random random;
    private int numRobots;
    private int numPlayers;
    private int rectSize;
    private Robot[][] robots;
    private int closestOpponent[];
    private int index;
    private int[][][] robotRectangles;
    private boolean robotGoesClockwise[];
    private int secretNumber = 0;
    private double paranoiaFactor = .3;  //the lower, the more paranoid
    private int[][] _board = null;
    
    public Robot[] register(Rectangles __rectangles) throws Exception
    {
        Robot[] RET;
        rect = __rectangles;
        if (random  == null) {
            random = new Random();
        }
	
        numRobots = rect.numRobots();
	numPlayers = rect.numPlayers();
        rectSize = rect.size();
	System.err.println("numRobots: " + numRobots + ", numPlayers: " + numPlayers + ", size: " + rectSize);//DEBUG
	closestOpponent = new int[numRobots];
	robotGoesClockwise = new boolean[numRobots];
	index = rect.indexOf(this);
	robotRectangles = new int[numRobots][2][2];  // [robot][vertex "1" or vertex "3"][X or Y]
	RET = new Robot[numRobots];
	secretNumber = (int)(Math.sqrt( ((double)(rectSize * rectSize)) / (((numPlayers - 1 > 0) ? (numPlayers - 1) : numPlayers) * (numRobots)) ) * paranoiaFactor);
	secretNumber = (secretNumber >= 3) ? secretNumber : 3;

	System.err.println("secret number: " + secretNumber);//DEBUG

	for (int i = 0; i < numRobots; i++) {
	    RET[i] = new Robot(random.nextInt(rectSize), random.nextInt(rectSize));
	    if(random.nextDouble() > 0.5)
		robotGoesClockwise[i] = true;
	    else
		robotGoesClockwise[i] = false;
	    rectangle(GET_NEW, RET[i].xpos(), RET[i].ypos(), i);
	}
        return RET;
    }
    
    public char[] move() throws Exception
    {
	char[] RET = new char[numRobots];
	int tmpDist = 0;
	robots = rect.allRobots();
	_board = rect.colors();
	
	for(int myRobot = 0; myRobot < numRobots; myRobot++)  //for each of our robots...
	    {
		closestOpponent[myRobot] = Integer.MAX_VALUE;
		for(int player = 0; player < numPlayers; player++)  ///for each of the players...
		    {
			if(player != index)  //(except ourselves)...
			    {
				for(int oppRobot = 0; oppRobot < numRobots; oppRobot++)  //for each of the player's robots...
				    {
					tmpDist = xyDistance(robots[player][oppRobot], robots[index][myRobot]);
					if(tmpDist < closestOpponent[myRobot])
					    closestOpponent[myRobot] = tmpDist;
				    }
			    }
		    }
		int[] start = getClosestPositionOnRectangle(robots[index][myRobot].xpos(),
							    robots[index][myRobot].ypos(),
							    robotRectangles[myRobot][VERTEX_1][X],
							    robotRectangles[myRobot][VERTEX_1][Y],
							    robotRectangles[myRobot][VERTEX_3][X],
							    robotRectangles[myRobot][VERTEX_3][Y]);
		tmpDist = movesToComplete(robots[index][myRobot].xpos(), robots[index][myRobot].ypos(), myRobot, start);
		System.err.println("robot" + myRobot + ", computed " + tmpDist + " moves to complete");//DEBUG
		if(start[X] == robots[index][myRobot].xpos() && start[Y] == robots[index][myRobot].ypos())
		    {//if robot is currently on square...
			if(tmpDist >= closestOpponent[myRobot] &&
			   robotRectangles[myRobot][VERTEX_3][Y] - robotRectangles[myRobot][VERTEX_1][Y] > 3 &&
			   robotRectangles[myRobot][VERTEX_3][X] - robotRectangles[myRobot][VERTEX_1][X] > 3)
			    {  //not enough time to finish, so short-circuit...
				System.err.println("robot" + myRobot + ", short-circuiting " + robotRectangles[myRobot][VERTEX_1][X] +
						   "," + robotRectangles[myRobot][VERTEX_1][Y] + "-" + robotRectangles[myRobot][VERTEX_3][X] +
						   "," + robotRectangles[myRobot][VERTEX_3][Y]);//DEBUG
				rectangle(SHORT_CIRCUIT, robots[index][myRobot].xpos(), robots[index][myRobot].ypos(), myRobot);
			    }
			else if(tmpDist == 0 || (tmpDist >= closestOpponent[myRobot] &&
						 (robotRectangles[myRobot][VERTEX_3][Y] - robotRectangles[myRobot][VERTEX_1][Y] <= 3 ||
						  robotRectangles[myRobot][VERTEX_3][X] - robotRectangles[myRobot][VERTEX_1][X] <= 3)))
			    {  //just finished forming a rectangle, start on something else...
				System.err.println("robot" + myRobot + ", replacing " + robotRectangles[myRobot][VERTEX_1][X] +
						   "," + robotRectangles[myRobot][VERTEX_1][Y] + "-" + robotRectangles[myRobot][VERTEX_3][X] +
						   "," + robotRectangles[myRobot][VERTEX_3][Y]);//DEBUG
				rectangle(GET_NEW, robots[index][myRobot].xpos(), robots[index][myRobot].ypos(), myRobot);
			    }
			else if(tmpDist < closestOpponent[myRobot])
			    {  //construct larger rectangle...
				System.err.println("robot" + myRobot + ", extending " + robotRectangles[myRobot][VERTEX_1][X] +
						   "," + robotRectangles[myRobot][VERTEX_1][Y] + "-" + robotRectangles[myRobot][VERTEX_3][X] +
						   "," + robotRectangles[myRobot][VERTEX_3][Y]);//DEBUG
				rectangle(EXTEND, robots[index][myRobot].xpos(), robots[index][myRobot].ypos(), myRobot);
			    }
		    }
		RET[myRobot] = getNextMove(myRobot, start);
		System.err.println("robot" + myRobot + ", moving " + RET[myRobot]);//DEBUG
	    }
        return RET;
    }
    
    public char getNextMove(int robotIndex, int[] closestPosOnRect) throws Exception
    {
	if(robotGoesClockwise[robotIndex])
	    {
		if(closestPosOnRect[X] != robots[index][robotIndex].xpos() ||
		   closestPosOnRect[Y] != robots[index][robotIndex].ypos())
		    {   //get robot to rectangle:
			if(robots[index][robotIndex].ypos() < robotRectangles[robotIndex][VERTEX_1][Y] &&
			   robots[index][robotIndex].xpos() >= robotRectangles[robotIndex][VERTEX_1][X])
			    return 'S';
			if(robots[index][robotIndex].xpos() > robotRectangles[robotIndex][VERTEX_3][X] &&
			   robots[index][robotIndex].ypos() >= robotRectangles[robotIndex][VERTEX_1][Y])
			    return 'W';
			if(robots[index][robotIndex].ypos() > robotRectangles[robotIndex][VERTEX_3][Y] &&
			   robots[index][robotIndex].xpos() <= robotRectangles[robotIndex][VERTEX_3][X])
			    return 'N';
			if(robots[index][robotIndex].xpos() < robotRectangles[robotIndex][VERTEX_1][X] &&
			   robots[index][robotIndex].ypos() <= robotRectangles[robotIndex][VERTEX_3][Y])
			    return 'E';
		    }
		
		//move robot along rectangle:
		if(robots[index][robotIndex].ypos() == robotRectangles[robotIndex][VERTEX_1][Y] &&
		   robots[index][robotIndex].xpos() != robotRectangles[robotIndex][VERTEX_3][X])
		    return 'E';
		if(robots[index][robotIndex].xpos() == robotRectangles[robotIndex][VERTEX_3][X] &&
		   robots[index][robotIndex].ypos() != robotRectangles[robotIndex][VERTEX_3][Y])
		    return 'S';
		if(robots[index][robotIndex].ypos() == robotRectangles[robotIndex][VERTEX_3][Y] &&
		   robots[index][robotIndex].xpos() != robotRectangles[robotIndex][VERTEX_1][X])
		    return 'W';
		if(robots[index][robotIndex].xpos() == robotRectangles[robotIndex][VERTEX_1][X] &&
		   robots[index][robotIndex].ypos() != robotRectangles[robotIndex][VERTEX_1][Y])
		    return 'N';
	    }
	else
	    {
		if(closestPosOnRect[X] != robots[index][robotIndex].xpos() ||
		   closestPosOnRect[Y] != robots[index][robotIndex].ypos())
		    {   //get robot to rectangle:
			if(robots[index][robotIndex].ypos() < robotRectangles[robotIndex][VERTEX_1][Y] &&
			   robots[index][robotIndex].xpos() <= robotRectangles[robotIndex][VERTEX_3][X])
			    return 'S';
			if(robots[index][robotIndex].xpos() > robotRectangles[robotIndex][VERTEX_3][X] &&
			   robots[index][robotIndex].ypos() <= robotRectangles[robotIndex][VERTEX_3][Y])
			    return 'W';
			if(robots[index][robotIndex].ypos() > robotRectangles[robotIndex][VERTEX_3][Y] &&
			   robots[index][robotIndex].xpos() >= robotRectangles[robotIndex][VERTEX_1][X])
			    return 'N';
			if(robots[index][robotIndex].xpos() < robotRectangles[robotIndex][VERTEX_1][X] &&
			   robots[index][robotIndex].ypos() >= robotRectangles[robotIndex][VERTEX_1][Y])
			    return 'E';
		    }
		
		if(robots[index][robotIndex].ypos() == robotRectangles[robotIndex][VERTEX_1][Y] &&
		   robots[index][robotIndex].xpos() != robotRectangles[robotIndex][VERTEX_1][X])
		    return 'W';
		if(robots[index][robotIndex].xpos() == robotRectangles[robotIndex][VERTEX_3][X] &&
		   robots[index][robotIndex].ypos() != robotRectangles[robotIndex][VERTEX_1][Y])
		    return 'N';
		if(robots[index][robotIndex].ypos() == robotRectangles[robotIndex][VERTEX_3][Y] &&
		   robots[index][robotIndex].xpos() != robotRectangles[robotIndex][VERTEX_3][X])
		    return 'E';
		if(robots[index][robotIndex].xpos() == robotRectangles[robotIndex][VERTEX_1][X] &&
		   robots[index][robotIndex].ypos() != robotRectangles[robotIndex][VERTEX_3][Y])
		    return 'S';
	    }
	//should never get here...
	return 'E';
    }
    
    public void rectangle(int behaviorCode, int x, int y, int robotIndex) throws Exception
    {
	int[] remainingSquares = new int[4];  //four sides: NW-NE, NE-SE, SE-SW, SW-NW
	System.err.println("robot" + robotIndex + ", entering rectangle: (" + robotRectangles[robotIndex][VERTEX_1][X] + "," + robotRectangles[robotIndex][VERTEX_1][Y] + "),(" + robotRectangles[robotIndex][VERTEX_3][X] + "," + robotRectangles[robotIndex][VERTEX_3][Y] + ")");//DEBUG
	switch(behaviorCode)
	    {
	    case SHORT_CIRCUIT:
		for(int i = robotRectangles[robotIndex][VERTEX_1][X]; i <= robotRectangles[robotIndex][VERTEX_3][X]; i++)
		    if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
			remainingSquares[0]++;
		for(int i = robotRectangles[robotIndex][VERTEX_1][Y]; i <= robotRectangles[robotIndex][VERTEX_3][Y]; i++)
		    if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
			remainingSquares[1]++;
		for(int i = robotRectangles[robotIndex][VERTEX_3][X]; i >= robotRectangles[robotIndex][VERTEX_1][X]; i--)
		    if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
			remainingSquares[2]++;
		for(int i = robotRectangles[robotIndex][VERTEX_3][Y]; i >= robotRectangles[robotIndex][VERTEX_1][Y]; i--)
		    if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
			remainingSquares[3]++;
		if(remainingSquares[0] >= remainingSquares[1] &&
		   remainingSquares[0] >= remainingSquares[2] &&
		   remainingSquares[0] >= remainingSquares[3] &&
		   robotRectangles[robotIndex][VERTEX_1][Y] < rectSize - 1 &&
		   y != robotRectangles[robotIndex][VERTEX_1][Y] &&
		   robotRectangles[robotIndex][VERTEX_3][Y] - robotRectangles[robotIndex][VERTEX_1][Y] >= 2)
		    robotRectangles[robotIndex][VERTEX_1][Y]++;
		else if(remainingSquares[1] >= remainingSquares[0] &&
			remainingSquares[1] >= remainingSquares[2] &&
			remainingSquares[1] >= remainingSquares[3] &&
			robotRectangles[robotIndex][VERTEX_3][X] > 0 &&
			x != robotRectangles[robotIndex][VERTEX_3][X] &&
			robotRectangles[robotIndex][VERTEX_3][X] - robotRectangles[robotIndex][VERTEX_1][X] >= 2)
		    robotRectangles[robotIndex][VERTEX_3][X]--;
		else if(remainingSquares[2] >= remainingSquares[0] &&
			remainingSquares[2] >= remainingSquares[1] &&
			remainingSquares[2] >= remainingSquares[3] &&
			robotRectangles[robotIndex][VERTEX_3][Y] > 0 &&
			y != robotRectangles[robotIndex][VERTEX_3][Y] &&
			robotRectangles[robotIndex][VERTEX_3][Y] - robotRectangles[robotIndex][VERTEX_1][Y] >= 2)
		    robotRectangles[robotIndex][VERTEX_3][Y]--;
		else if(remainingSquares[3] >= remainingSquares[0] &&
			remainingSquares[3] >= remainingSquares[1] &&
			remainingSquares[3] >= remainingSquares[2] &&
			robotRectangles[robotIndex][VERTEX_1][X] < rectSize - 1 &&
			x != robotRectangles[robotIndex][VERTEX_1][X] &&
			robotRectangles[robotIndex][VERTEX_3][X] - robotRectangles[robotIndex][VERTEX_3][X] >= 2)
		    robotRectangles[robotIndex][VERTEX_1][X]++;
		break;
	    case GET_NEW:
		int tmpX, tmpY;
		/*     		int[] score = {0, 0, 0, 0};
		for(int i = ((x - secretNumber) >= 0 ? x - secretNumber : 0); i < x; i++)
		    for(int j = ((y - secretNumber) >= 0 ? (y - secretNumber) : 0); j < y; j++)
			{
			    if(_board[i][j] == index)
				score[0] += 2;
			    else if(_board[i][j] == Rectangles._CFILLED)
				score[0] -= 3;
			    else
				score[0]++;
			}
		for(int i = x; i < ((x + secretNumber) < rectSize ? x + secretNumber : rectSize); i++)
		    for(int j = ((y - secretNumber) >= 0 ? (y - secretNumber) : 0); j < y; j++)
			{
			    if(_board[i][j] == index)
				score[1] += 2;
			    else if(_board[i][j] == Rectangles._CFILLED)
				score[1] -= 3;
			    else
				score[1]++;
			}
		for(int i = ((x - secretNumber) >= 0 ? x - secretNumber : 0); i < x; i++)
		    for(int j = y; j < ((y + secretNumber) < rectSize ? y + secretNumber: rectSize); j++)
			{
			    if(_board[i][j] == index)
				score[2] += 2;
			    else if(_board[i][j] == Rectangles._CFILLED)
				score[2] -= 3;
			    else
				score[2]++;
			}
		for(int i = x; i < ((x + secretNumber) < rectSize ? x + secretNumber : rectSize); i++)
		    for(int j = y; j < ((y + secretNumber) < rectSize ? y + secretNumber : rectSize); j++)
			{
			    if(_board[i][j] == index)
				score[3] += 2;
			    else if(_board[i][j] == Rectangles._CFILLED)
				score[3] -= 3;
			    else
				score[3]++;
			}
		if(score[0] <= score[1] && score[0] <= score[2] && score[0] <= score[3])
		    {
			tmpX = (x - secretNumber) >= 0 ? x - secretNumber : 0;
			tmpY = (y - secretNumber) >= 0 ? y - secretNumber : 0;
		    }
		else if(score[1] <= score[0] && score[1] <= score[2] && score[1] <= score[3])
		    {
			tmpX = x;
			tmpY = (y - secretNumber) >= 0 ? y - secretNumber : 0;
		    }
		else if(score[2] <= score[0] && score[2] <= score[1] && score[2] <= score[3])
		    {
			tmpX = (x - secretNumber) >= 0 ? x - secretNumber : 0;
			tmpY = y;
		    }
		else
		{*/
			tmpX = x;
			tmpY = y;
			/*		    }*/
		robotRectangles[robotIndex][VERTEX_1][X] = tmpX % (rectSize - secretNumber + 1);
		robotRectangles[robotIndex][VERTEX_1][Y] = tmpY % (rectSize - secretNumber + 1);
		robotRectangles[robotIndex][VERTEX_3][X] = robotRectangles[robotIndex][VERTEX_1][X] + secretNumber - 1;
		robotRectangles[robotIndex][VERTEX_3][Y] = robotRectangles[robotIndex][VERTEX_1][Y] + secretNumber - 1;
		break;
	    case EXTEND:
		for(int i = robotRectangles[robotIndex][VERTEX_1][X]; i <= robotRectangles[robotIndex][VERTEX_3][X]; i++)
		    if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
			remainingSquares[0]++;
		for(int i = robotRectangles[robotIndex][VERTEX_1][Y]; i <= robotRectangles[robotIndex][VERTEX_3][Y]; i++)
		    if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
			remainingSquares[1]++;
		for(int i = robotRectangles[robotIndex][VERTEX_3][X]; i >= robotRectangles[robotIndex][VERTEX_1][X]; i--)
		    if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
			remainingSquares[2]++;
		for(int i = robotRectangles[robotIndex][VERTEX_3][Y]; i >= robotRectangles[robotIndex][VERTEX_1][Y]; i--)
		    if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
			remainingSquares[3]++;
		if(remainingSquares[0] >= remainingSquares[1] &&
		   remainingSquares[0] >= remainingSquares[2] &&
		   remainingSquares[0] >= remainingSquares[3] &&
		   robotRectangles[robotIndex][VERTEX_1][Y] > 0 &&
		   y != robotRectangles[robotIndex][VERTEX_1][Y])
		    robotRectangles[robotIndex][VERTEX_1][Y]--;
		else if(remainingSquares[1] >= remainingSquares[0] &&
			remainingSquares[1] >= remainingSquares[2] &&
			remainingSquares[1] >= remainingSquares[3] &&
			robotRectangles[robotIndex][VERTEX_3][X] < rectSize - 1 &&
			x != robotRectangles[robotIndex][VERTEX_3][X])
		    robotRectangles[robotIndex][VERTEX_3][X]++;
		else if(remainingSquares[2] >= remainingSquares[0] &&
			remainingSquares[2] >= remainingSquares[1] &&
			remainingSquares[2] >= remainingSquares[3] &&
			robotRectangles[robotIndex][VERTEX_3][Y] < rectSize - 1 &&
			y != robotRectangles[robotIndex][VERTEX_3][Y])
		    robotRectangles[robotIndex][VERTEX_3][Y]++;
		else if(remainingSquares[3] >= remainingSquares[0] &&
			remainingSquares[3] >= remainingSquares[1] &&
			remainingSquares[3] >= remainingSquares[2] &&
			robotRectangles[robotIndex][VERTEX_1][X] > 0 &&
			x != robotRectangles[robotIndex][VERTEX_1][X])
		    robotRectangles[robotIndex][VERTEX_1][X]--;
		break;
	    }
	System.err.println("robot" + robotIndex + ", leaving rectangle: (" + robotRectangles[robotIndex][VERTEX_1][X] + "," + robotRectangles[robotIndex][VERTEX_1][Y] + "),(" + robotRectangles[robotIndex][VERTEX_3][X] + "," + robotRectangles[robotIndex][VERTEX_3][Y] + ")");//DEBUG
    }
    
    public int xyDistance(Robot r1, Robot r2) throws Exception
    {
	return Math.abs(r1.xpos() - r2.xpos()) + Math.abs(r1.ypos() - r2.ypos());
    }

    public int[] getClosestPositionOnRectangle(int x, int y, int v1x, int v1y, int v3x, int v3y)
    {
	int[] start = new int[2];

	start[X] = x;
	start[Y] = y;
	if(x <= v1x)
	    {
		if(y <= v1y)
		    {
			start[X] = v1x;
			start[Y] = v1y;
		    }
		else if(y >= v3y)
		    {
			start[X] = v1x;
			start[Y] = v3y;
		    }
		else
		    start[X] = v1x;
	    }
	else if(x >= v3x)
	    {
		if(y <= v1y)
		    {
			start[X] = v3x;
			start[Y] = v1y;
		    }
		else if(y >= v3y)
		    {
			start[X] = v3x;
			start[Y] = v3y;
		    }
		else
		    start[X] = v3x;
	    }
	else
	    {
		if(y <= v1y)
		    start[Y] = v1y;
		else if(y >= v3y)
		    start[Y] = v3y;
		else
		    {
			if(y - v1y <= v3y - y)
			    {
				if(x - v1x < y - v1y)
				    start[X] = v1x;
				else if(v3x - x < y - v1y)
				    start[X] = v3x;
				else
				    start[Y] = v1y;
			    }
			else
			    {
				if(x - v1x < v3y - y)
				    start[X] = v1x;
				else if(v3x - x < v3y - y)
				    start[X] = v3x;
				else
				    start[Y] = v3y;
			    }
		    }
	    }
	return start;
    }
    
    //should be adapted to take into account "filled" squares blocking its path
    public int movesToComplete(int x, int y, int robotIndex, int[] start)
    {
	int startX, startY;  //the position from which to start counting ON THE RECTANGLE
	int moves = 0;
	int movesSinceLastBlankSquare = 0;
	moves += Math.abs(x - start[X]) + Math.abs(y - start[Y]);  //now equals moves necessary to reach rectangle

	System.err.println("robot" + robotIndex + ", movesToComplete:' (" + robotRectangles[robotIndex][VERTEX_1][X] + "," + robotRectangles[robotIndex][VERTEX_1][Y] + "),(" + robotRectangles[robotIndex][VERTEX_3][X] + "," + robotRectangles[robotIndex][VERTEX_3][Y] + ")");//DEBUG


	//first check if rectangle can never be completed:
	/*	boolean blocked = false;
	System.err.println("size: " + _board.length + ", 1x: " + robotRectangles[robotIndex][VERTEX_1][X] + ", 1y: " + robotRectangles[robotIndex][VERTEX_1][Y] + ", 3x: " + robotRectangles[robotIndex][VERTEX_3][X] + ", 3y: " + robotRectangles[robotIndex][VERTEX_3][Y]);//DEBUG
	for(int i = robotRectangles[robotIndex][VERTEX_1][X]; i <= robotRectangles[robotIndex][VERTEX_3][X] && !blocked; i++)
	    {
		for(int j = robotRectangles[robotIndex][VERTEX_1][Y]; j <= robotRectangles[robotIndex][VERTEX_3][Y] && !blocked; j++)
		    {
			if(_board[i][j] == _CFILLED)
			    blocked = true;
		    }
	    }
	if(blocked)
	    return Integer.MAX_VALUE;
	*/
	if(start[X] > robotRectangles[robotIndex][VERTEX_1][X] && start[X] < robotRectangles[robotIndex][VERTEX_3][X])
	    {  //moving W <-> E
		if(robotGoesClockwise[robotIndex])
		    {
			if(start[Y] == robotRectangles[robotIndex][VERTEX_1][Y])
			    {  //moving E along N-most horizontal
				for(int i = start[X] + 1; i <= robotRectangles[robotIndex][VERTEX_3][X]; i++)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;
				    }
				//now moving S along E-most side
				for(int i = start[Y] + 1; i <= robotRectangles[robotIndex][VERTEX_3][Y]; i++)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//now moving W along S-most side
				for(int i = robotRectangles[robotIndex][VERTEX_3][X] - 1; i >= robotRectangles[robotIndex][VERTEX_1][X]; i--)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;
				    }
				//now moving N along W-most side
				for(int i = robotRectangles[robotIndex][VERTEX_3][Y] - 1; i >= robotRectangles[robotIndex][VERTEX_1][Y]; i--)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;
				    }
				//now moving E along N-most horizontal (again)
				for(int i = robotRectangles[robotIndex][VERTEX_1][X] + 1; i < start[X]; i++)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;
				    }
			    }
			else
			    {  //moving W along S-most horizontal
				for(int i = start[X] - 1; i >= robotRectangles[robotIndex][VERTEX_1][X]; i--)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//now moving N along W-most side
				for(int i = robotRectangles[robotIndex][VERTEX_3][Y] - 1; i >= robotRectangles[robotIndex][VERTEX_1][Y]; i--)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;
				    }
				//now moving E along N-most horizontal
				for(int i = robotRectangles[robotIndex][VERTEX_1][X] + 1; i < robotRectangles[robotIndex][VERTEX_3][X]; i++)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;
				    }
				//now moving S along E-most side
				for(int i = robotRectangles[robotIndex][VERTEX_1][Y] + 1; i <= robotRectangles[robotIndex][VERTEX_3][Y]; i++)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//now moving W along S-most horizontal (again)
				for(int i = robotRectangles[robotIndex][VERTEX_3][X] - 1; i > start[X]; i--)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
			    }
		    }
		else
		    {  //count around the rectangle (counterclockwise) for blank spots
			if(start[Y] == robotRectangles[robotIndex][VERTEX_1][Y])
			    {  //moving W along N-most horizontal
				for(int i = start[X] - 1; i >= robotRectangles[robotIndex][VERTEX_1][X]; i--)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//now moving S along W-most side
				for(int i = start[Y] + 1; i <= robotRectangles[robotIndex][VERTEX_3][Y]; i++)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//now moving E along S-most horizontal
				for(int i = robotRectangles[robotIndex][VERTEX_1][X] + 1; i <= robotRectangles[robotIndex][VERTEX_3][X]; i++)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//now moving N along E-most side
				for(int i = robotRectangles[robotIndex][VERTEX_3][Y] - 1; i >= start[Y]; i--)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//now moving W along N-most horizontal (again)
				for(int i = robotRectangles[robotIndex][VERTEX_3][X] - 1; i > start[X]; i--)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
			    }
			else
			    {
				//moving E along S-most horizontal
				for(int i = start[X]; i <= robotRectangles[robotIndex][VERTEX_3][X]; i++)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//now moving N along E-most side
				for(int i = start[Y] - 1; i >= robotRectangles[robotIndex][VERTEX_1][Y]; i--)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//now moving W along N-most horizontal
				for(int i = robotRectangles[robotIndex][VERTEX_3][X] - 1; i >= robotRectangles[robotIndex][VERTEX_1][X]; i--)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//now moving S along W-most horizontal
				for(int i = robotRectangles[robotIndex][VERTEX_1][Y] + 1; i <= start[Y]; i++)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//now moving E along S-most horizontal (again)
				for(int i = robotRectangles[robotIndex][VERTEX_1][X] + 1; i < start[X]; i++)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
			    }
		    }
	    }
	else if(start[Y] > robotRectangles[robotIndex][VERTEX_1][Y] && start[Y] < robotRectangles[robotIndex][VERTEX_3][Y])
	    {  //moving N <-> S
		if(robotGoesClockwise[robotIndex])
		    {
			if(start[X] == robotRectangles[robotIndex][VERTEX_1][X])
			    {  //moving N on W-most side
				for(int i = start[Y] - 1; i >= robotRectangles[robotIndex][VERTEX_1][Y]; i--)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//moving E on N-most horizontal
				for(int i = start[X] + 1; i <= robotRectangles[robotIndex][VERTEX_3][X]; i++)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//moving S on E-most side
				for(int i = robotRectangles[robotIndex][VERTEX_1][Y] + 1; i <= robotRectangles[robotIndex][VERTEX_3][Y]; i++)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//moving W on S-most horizontal
				for(int i = robotRectangles[robotIndex][VERTEX_3][X] - 1; i >= start[X]; i--)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//moving N on W-most side (again)
				for(int i = robotRectangles[robotIndex][VERTEX_3][Y] - 1; i > start[Y]; i--)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
			    }
			else
			    {  //moving S on E-most side
				for(int i = start[Y] + 1; i <= robotRectangles[robotIndex][VERTEX_3][Y]; i++)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//moving W on S-most horizontal
				for(int i = start[X] - 1; i >= robotRectangles[robotIndex][VERTEX_1][X]; i--)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//moving N on W-most side
				for(int i = robotRectangles[robotIndex][VERTEX_3][Y] - 1; i >= robotRectangles[robotIndex][VERTEX_1][Y]; i--)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//moving E on N-most horizontal
				for(int i = robotRectangles[robotIndex][VERTEX_1][X] + 1; i <= start[X]; i++)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;					    
				    }
				//moving S on E-most side (again)
				for(int i = robotRectangles[robotIndex][VERTEX_1][Y] + 1; i < start[Y]; i++)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;				    
				    }
			    }
		    }
		else  //counterclockwise
		    {
			if(start[X] == robotRectangles[robotIndex][VERTEX_1][X])
			    {  //moving S on W-most side
				for(int i = start[Y] + 1; i <= robotRectangles[robotIndex][VERTEX_3][Y]; i++)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;				    
				    }
				//moving E on S-most horizontal
				for(int i = start[X] + 1; i <= robotRectangles[robotIndex][VERTEX_3][X]; i++)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;				    
				    }
				//moving N on E-most side
				for(int i = robotRectangles[robotIndex][VERTEX_3][Y] - 1; i >= robotRectangles[robotIndex][VERTEX_1][Y]; i--)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;				    
				    }
				//moving W on N-most horizontal
				for(int i = robotRectangles[robotIndex][VERTEX_3][X] - 1; i >= start[X]; i--)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_1][X]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;				    
				    }
				//moving S on W-most side (again)
				for(int i = robotRectangles[robotIndex][VERTEX_1][Y] + 1; i < start[Y]; i++)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;				    
				    }
			    }
			else
			    {  //moving N on E-most side
				for(int i = start[Y] - 1; i >= robotRectangles[robotIndex][VERTEX_1][Y]; i--)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;
				    }
				//moving W on N-most horizontal
				for(int i = start[X] - 1; i >= robotRectangles[robotIndex][VERTEX_1][X]; i--)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;
				    }
				//moving S on W-most side
				for(int i = robotRectangles[robotIndex][VERTEX_1][Y] + 1; i <= robotRectangles[robotIndex][VERTEX_3][Y]; i++)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;
				    }
				//moving E on S-most horizontal
				for(int i = robotRectangles[robotIndex][VERTEX_1][X] + 1; i <= start[X]; i++)
				    {
					moves++;
					if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;
				    }
				//moving N on E-most side (again)
				for(int i = robotRectangles[robotIndex][VERTEX_3][Y] - 1; i > start[Y]; i--)
				    {
					moves++;
					if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
					    movesSinceLastBlankSquare = 0;
					else
					    movesSinceLastBlankSquare++;
				    }
			    }			
		    }
	    }
	else  //corners:
	    {
		if(robotGoesClockwise[robotIndex])
		    {
			if(start[Y] == robotRectangles[robotIndex][VERTEX_1][Y])  //upper two corners
			    {
				if(start[X] == robotRectangles[robotIndex][VERTEX_1][X])  //upper left corner
				    {
					//E along N-most side
					for(int i = start[X] + 1; i <= robotRectangles[robotIndex][VERTEX_3][X]; i++)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//S along E-most side
					for(int i = start[Y] + 1; i <= robotRectangles[robotIndex][VERTEX_3][Y]; i++)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//W along S-most side
					for(int i = robotRectangles[robotIndex][VERTEX_3][X] - 1; i >= start[X]; i--)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//N along W-most side
					for(int i = robotRectangles[robotIndex][VERTEX_3][Y] - 1; i > start[Y]; i--)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
				    }
				else  //upper right corner
				    {
					//S along E-most side
					for(int i = start[Y] + 1; i <= robotRectangles[robotIndex][VERTEX_3][Y]; i++)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//W along S-most side
					for(int i = start[X] - 1; i >= robotRectangles[robotIndex][VERTEX_1][X]; i--)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//N along W-most side
					for(int i = robotRectangles[robotIndex][VERTEX_3][Y] - 1; i >= start[Y]; i--)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//E along N-most side
					for(int i = robotRectangles[robotIndex][VERTEX_1][X] + 1; i < start[X]; i++)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
				    }
			    }

			else
			    {  //lower two corners
				if(start[X] == robotRectangles[robotIndex][VERTEX_1][X])  //lower left corner
				    {
					//N along W-most side
					for(int i = start[Y] - 1; i >= robotRectangles[robotIndex][VERTEX_1][Y]; i--)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//E along N-most side
					for(int i = start[X] + 1; i <= robotRectangles[robotIndex][VERTEX_3][X]; i++)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//S along E-most side
					for(int i = robotRectangles[robotIndex][VERTEX_1][Y] + 1; i <= start[Y]; i ++)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//W along S-most side
					for(int i = robotRectangles[robotIndex][VERTEX_3][X] - 1; i > start[X]; i--)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
				    }
				else  //lower right corner
				    {
					//W along S-most side
					for(int i = start[X] - 1; i >= robotRectangles[robotIndex][VERTEX_1][X]; i--)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//N along W-most side
					for(int i = start[Y] - 1; i >= robotRectangles[robotIndex][VERTEX_1][Y]; i--)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//E along N-most side
					for(int i = robotRectangles[robotIndex][VERTEX_1][X] + 1; i <= start[X]; i++)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//S along E-most side
					for(int i = robotRectangles[robotIndex][VERTEX_1][Y] + 1; i < start[Y]; i++)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
				    }
			    }
		    }
		else  //counterclockwise
		    {
			if(start[Y] == robotRectangles[robotIndex][VERTEX_1][Y])  //upper two corners
			    {
				if(start[X] == robotRectangles[robotIndex][VERTEX_1][X])  //upper left corner
				    {
					//S along W-most side
					for(int i = start[Y] + 1; i <= robotRectangles[robotIndex][VERTEX_3][Y]; i++)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//E along S-most side
					for(int i = start[X] + 1; i <= robotRectangles[robotIndex][VERTEX_3][X]; i++)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//N along E-most side
					for(int i = robotRectangles[robotIndex][VERTEX_3][Y] - 1; i >= start[Y]; i--)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//W along N-most side
					for(int i = robotRectangles[robotIndex][VERTEX_3][X] - 1; i > start[X]; i--)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
				    }
				else  //upper right corner
				    {
					//W along N-most side
					for(int i = start[X] - 1; i >= robotRectangles[robotIndex][VERTEX_1][X]; i--)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//S along W-most side
					for(int i = start[Y] + 1; i <= robotRectangles[robotIndex][VERTEX_3][Y]; i++)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//E along S-most side
					for(int i = robotRectangles[robotIndex][VERTEX_1][X] + 1; i <= start[X]; i++)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//N along E-most side
					for(int i = robotRectangles[robotIndex][VERTEX_3][Y] - 1; i > start[Y]; i--)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
				    }
			    }
			else
			    {  //lower two corners
				if(start[X] == robotRectangles[robotIndex][VERTEX_1][X])  //lower left corner
				    {
					//E along S-most side
					for(int i = start[X] + 1; i <= robotRectangles[robotIndex][VERTEX_1][X]; i++)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//N along E-most side
					for(int i = start[Y] - 1; i >= robotRectangles[robotIndex][VERTEX_1][Y]; i--)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//W along N-most side
					for(int i = robotRectangles[robotIndex][VERTEX_3][X] - 1; i >= start[X]; i--)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//S along W-most side
					for(int i = robotRectangles[robotIndex][VERTEX_1][Y] + 1; i < start[Y]; i++)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
				    }
				else  //lower right corner
				    {
					//N along E-most side
					for(int i = start[Y] - 1; i >= robotRectangles[robotIndex][VERTEX_1][Y]; i--)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_3][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//W along N-most side
					for(int i = start[X] - 1; i >= robotRectangles[robotIndex][VERTEX_1][X]; i--)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_1][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//S along W-most side
					for(int i = robotRectangles[robotIndex][VERTEX_1][Y] + 1; i <= start[Y]; i++)
					    {
						moves++;
						if(_board[robotRectangles[robotIndex][VERTEX_1][X]][i] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
					//E along S-most side
					for(int i = robotRectangles[robotIndex][VERTEX_1][X] + 1; i < start[X]; i++)
					    {
						moves++;
						if(_board[i][robotRectangles[robotIndex][VERTEX_3][Y]] != index)
						    movesSinceLastBlankSquare = 0;
						else
						    movesSinceLastBlankSquare++;
					    }
				    }
			    }
		    }
	    }
	moves -= movesSinceLastBlankSquare;

	return moves;
    }

    public String name() throws Exception {
        return _CNAME;
    }

    public Color color() throws Exception {
        return _CCOLOR;
    }

    public boolean interactive() throws Exception {
        return false;
    }
}
